from time import sleep, time
from machine import Pin, PWM, Timer

print("Application: Water Tank Level Alarm - ver4.11 3 April 2021")

'''
Water Tank Level Alarm and Power Controller

Author: Steven Mulholland
Date: 4 April 2021
Version:  4.11

Revisions:
Version 4.11
1. Set buzzer and process durations to production levels
2. Tidied up the code and added some comments.
3. Added a power-up read of sensor lines to reset the alarm/warning flags.
4. Added a slow 2 flash of the Status LED on power up to indicate main.py was started ok.

Version 4:
1. Put Relay pin back to D1 to match the relay shield.
2. Modified alarm() function to set faster Alarm Led flashing.
    Also turned off the warning Led when alarm led flashing.

Version 3:
1. Setup a Process Delay (sleep(process_delay)) to only check the flags once per period.
    For testing that can be 5 seconds.  Production I will start with 30-60 seconds.
    Alternatively put the main process loop into a function and call it with a
    periodic hardware timer every 60 seconds.  Test with simple sleep() first. Works.

Version 2:
1. Now works.  Needed the global declaration inside ISR to support writing
   out to external variables.  Different from standard functions.
2. Idiot note re buzzer:  Don't name a function the same as a Pin object!!
3. Implemented hardware Timer + lambda function to turn off the buzzer - WORKS GREAT

Desciption
~~~~~~~~~~
Purpose
    1. To monitor and display water tank levels.
    2. If water level drops below the low level,
       cut power to water pumps to prevent the tank from getting completely empty.

Uses 2 x induction water sensors on the home water tank.
    1. Warning Sensor placed at 60% level
    2. Alarm sensor placed at 25% level
    
    The sensors I used are NPN (XKC-Y25-NPN) which means the Output lead is always a SINK.
    This needs to be converted to a +3.3V SOURCE to feed the sensor GPIO pins INPUT.
    Use optocouplers (PC817) for this.
    The sensors have a Mode lead which sets which condition connects the Outlput lead to GND.
    I have set this to connect to GND when there is no induction (i.e. water NOT detected).
    This allows current to flow into an Optocoupler which turns ON a +3.3V signal to the
    sensor GPIO Pin indicating a "No Water" event.

Indicators:
    1. Status LED (Green)  lights up solid when code running
    2. Warning LED (Orange) blinks 1 x per second when warning sensor triggered
       Stays flashing until Warning level is back above 60%
    3. Alarm LED (Red) blinks 2 times per second when alarm level reached
       Stays flashing until Warning level is back above 25%
    4. Buzzer.  Activated at both Warning and Alarm levels.

Power:
Dual AC Mains sockets for:
i) House Water Pump
ii) RO Filter pump
Relay connected to water pump mains supply.  Uses the Normally Closed (NC)
connection so that if controller power is OFF the power is ON to the pumps.
When the Alarm level is triggered the relay is ENABLED which cuts power to the AC and pumps

Switches:
1.  AC Mains Switch
    Cuts ALL power to controller AND AC sockets (pumps)
2.  MCU Power Switch - Override
    Connected to the output of the +5V power supply going to the MCU
    This halts the MCU, all LEDs shutoff and the relay is DISABLED which turns the
    power back ON to the pumps
    This allows a manual override to ALarm so the water can be turned back on when the
    level is still below 25%
    Note:  Power is still ENABLED to the Water Level Sensors which both have external LEDs.

'''

# Set the timing durations
warning_buzzer_duration = 5  # seconds for warning buzzer
alarm_buzzer_duration = 15    # seconds for alarm buzzer                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              # seconds
process_delay = 30            #testing = 5, production = 60.  Only checks sesnors once per this period.

status_led = Pin(16, Pin.OUT)    # GPIO16 D0
warning_led = PWM(Pin(14))       # GPIO14 D5 
alarm_led = PWM(Pin(15))         # GPIO15 D8 - Pulled to GND
relay = Pin(5, Pin.OUT)          # GPIO5  D1 (Wemos Relay Shield GPIO Pin)
warning_sensor = Pin(12, Pin.IN) # GPIO12 D6
alarm_sensor = Pin(13, Pin.IN)   # GPIO13 D7
buzzer = Pin(4, Pin.OUT)         # GPIO4  D2

# initialise GPIO Pins
relay.off()         # relay OFF = Power ON to pumps
buzzer.off()        # Warning buzzer OFF
warning_led.duty(0) # warning LED (Orange)
warning_led.freq(1) # once per second
alarm_led.duty(0)   # alarm LED and power cutoff (Red)
alarm_led.freq(2)   # twice per second

# Enable Status LED and flash it once on startup
status_led.on()
sleep(1)
status_led.off()
sleep(1)
status_led.on()
sleep(1)
status_led.off()
sleep(1)
status_led.on()

t1 = Timer(0)       # buzzer timer

# Reset flags to match sensors on power-up
if warning_sensor.value() == 0:
    warning_sensor_flag = False
else:
    warning_sensor_flag = True
if alarm_sensor.value() == 0:
    alarm_sensor_flag = False
else:
    alarm_sensor_flag = True


# Function to set sensor flags on interrupt trigger (rising and falling)
# Note:  must have the global declaration to allow saving results
def on_sensor_trigger(pin):
    global warning_sensor_flag
    global alarm_sensor_flag
    
    if warning_sensor.value() == 1:
        warning_sensor_flag = True
    else:
        warning_sensor_flag = False
    if alarm_sensor.value() == 1:
        alarm_sensor_flag = True
    else:
        alarm_sensor_flag = False
    print("Sensor Triggered on pin : ", pin)
    print("Warning sensor : ", warning_sensor_flag)
    print("Alarm sensor : ", alarm_sensor_flag)

# Turn on the buzzer then set a hardware timer to turn it off after duration time.
def buzzer_on(duration):
    buzzer.on()
    t1.init(period=duration*1000, mode=Timer.ONE_SHOT, callback=lambda t:buzzer.off())


# Primary alarm action logic.  Actions based on 'level' argument command.
def alarm(level):
    print("Alarm Level : ", level)
    if level == "Warning":
        # sound buzzer for one minute & set warning led to blinking
        warning_led.freq(1)
        warning_led.duty(512)
        buzzer_on(warning_buzzer_duration)
        relay.off()
        
    if level == "Alarm":
        # Enable alarm LED & sound buzzer for 3 minutes & ON the relay
        warning_led.duty(0)
        alarm_led.freq(3)
        alarm_led.duty(512)
        relay.on()  # Enable relay to DISABLE power to the pumps (NC connection on relay
        buzzer_on(alarm_buzzer_duration)
        
    if level == "ClearWarning":
        # disable warning LED
        warning_led.duty(0)
        relay.off()
        
    if level == "ClearAlarm":
        # disable Alarm LED and turn OFF the relay
        relay.off()
        alarm_led.duty(0)


# Interrupts.  default is triggered on Rising & Falling
warning_sensor.irq(handler=on_sensor_trigger)
alarm_sensor.irq(handler=on_sensor_trigger)

# main code here ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# flags for warning and alarm when currently set
alarm_triggered = False
warning_triggered = False

while True:
    # if warning not already set and top sensor only => warning
    if not warning_triggered and warning_sensor_flag and not alarm_sensor_flag:
        print("WARNNG >>>>")
        alarm("Warning")
        warning_triggered = True
        
    # if alarm not already set and bottom sensor => alarm & cut power    
    if not alarm_triggered and alarm_sensor_flag:
        print("ALARM >>>>")
        alarm("Alarm")
        alarm_triggered = True
    
    # reset alarm when lower sensor has water.
    if alarm_triggered and not alarm_sensor_flag:
        # An alarm has been triggered but now water level is ok at low levels
        # clear the alarm and enable power
        print("CLEAR ALARM >>>>")
        alarm_triggered = False
        alarm("ClearAlarm")
        alarm("Warning")
        warning_triggered = True
        
    # reset warning when top sensor has water
    if warning_triggered and not warning_sensor_flag:
        # A warning has been triggered but now water level is ok at high level
        # clear the warning
        print("CLEAR WARNING >>>>")
        warning_triggered = False
        alarm("ClearWarning")
    
    sleep(process_delay)


